module noise (
    input             in_apu_clk,
    input             in_qfr_clk,
    input             in_hfr_clk,
    input      [ 7:0] reg_0,
    input      [ 7:0] reg_1,
    input      [ 7:0] reg_2,
    output reg signed [15:0] noise_out
);

    // reg_0
    wire       lc_halt;
    wire       const_vol;
    wire [3:0] env_prd;
    
    assign lc_halt    = reg_0[5];
    assign const_vol  = reg_0[4];
    assign env_prd    = reg_0[3:0];
    
    // reg_1
    wire       mode;
    wire [3:0] prd_raw;

    assign mode    = reg_1[7];
    assign prd_raw = reg_1[3:0];

    // reg_2
    wire [4:0] lc_load;

    assign lc_load = reg_2[7:3];

    /// Envelope unit /////////////////////////////////////////////////////////
    reg [ 3:0] env_div;
    reg [ 3:0] env_cnt;
    reg [ 3:0] env_out;
    reg [23:0] env_oR;
    reg        env_start;

    always @( posedge in_qfr_clk ) begin

        // Reset
        if ( env_start ) begin
            env_start <= 1'b0;
            env_cnt   <= 4'hf;
            env_div   <= env_prd;
            env_oR    <= {reg_2, reg_1, reg_0};
        end

        else begin

            // Decrement counter when clocked by divider
            if ( env_div == 4'h0 ) begin
                env_div <= env_prd;

                if ( env_cnt > 4'h0 ) begin
                    env_cnt <= env_cnt - 4'h1;
                end

                // If loop flag is set, reload counter
                else if ( lc_halt ) begin
                    env_cnt <= 4'hf;
                end
            end

            // Decrement divider
            else begin
                env_div <= env_div - 4'h1;
            end

            // Volume output depends on constant volume flag
            if ( const_vol ) begin
                env_out <= env_prd;
            end
            else begin
                env_out <= env_cnt;
            end

            // Watch for reset
            if ( env_oR != {reg_2, reg_1, reg_0} ) begin
                env_start <= 1'b1;
            end

        end // end else block

    end // end always

    /// Timer & shift register ////////////////////////////////////////////////
    reg [11:0] prd;
    
    always @( * )
    begin
        case ( prd_raw )
            4'h0: prd <= 11'd4;
            4'h1: prd <= 11'd8;
            4'h2: prd <= 11'd16;
            4'h3: prd <= 11'd32;
            4'h4: prd <= 11'd64;
            4'h5: prd <= 11'd96;
            4'h6: prd <= 11'd128;
            4'h7: prd <= 11'd160;
            4'h8: prd <= 11'd202;
            4'h9: prd <= 11'd254;
            4'hA: prd <= 11'd380;
            4'hB: prd <= 11'd508;
            4'hC: prd <= 11'd762;
            4'hD: prd <= 11'd1016;
            4'hE: prd <= 11'd2034;
            4'hF: prd <= 11'd4068;
        endcase
    end

    reg [14:0] shift_reg = 15'd1;
	reg        feedback;
    reg [11:0] t;
	reg        shift_reset;

    always @( posedge in_apu_clk ) begin

        // When clocked by timer
        if ( t == 12'd0 ) begin

            // Calculate feedback
            if ( mode ) begin
                feedback = shift_reg[0] ^ shift_reg[6];
            end
            else begin
                feedback = shift_reg[0] ^ shift_reg[1];
            end

            // Shift right
            shift_reg = shift_reg >> 1;

            // Set leftmost bit to feedback
            shift_reg[14] = feedback;

            if ( shift_reg[0] || lc == 8'd0 ) begin
                noise_out <= 16'h0000;
            end
            else begin
                noise_out <=  (env_out << 11);
            end

            // Reload timer
            t <= prd;
        end

        // Decrement timer
        else begin
            t <= t - 12'd1;
        end

    end // end always


    /// Length counter ////////////////////////////////////////////////////////
    reg [ 7:0] lc_entry;
    
    always @( * )
    begin
        case ( lc_load )
            5'h00: lc_entry <= 8'h0A;
            5'h01: lc_entry <= 8'hFE;
            5'h02: lc_entry <= 8'h14;
            5'h03: lc_entry <= 8'h02;
            5'h04: lc_entry <= 8'h28;
            5'h05: lc_entry <= 8'h04;
            5'h06: lc_entry <= 8'h50;
            5'h07: lc_entry <= 8'h06;
            5'h08: lc_entry <= 8'hA0;
            5'h09: lc_entry <= 8'h08;
            5'h0A: lc_entry <= 8'h3C;
            5'h0B: lc_entry <= 8'h0A;
            5'h0C: lc_entry <= 8'h0E;
            5'h0D: lc_entry <= 8'h0C;
            5'h0E: lc_entry <= 8'h1A;
            5'h0F: lc_entry <= 8'h0E;
            5'h10: lc_entry <= 8'h0C;
            5'h11: lc_entry <= 8'h10;
            5'h12: lc_entry <= 8'h18;
            5'h13: lc_entry <= 8'h12;
            5'h14: lc_entry <= 8'h30;
            5'h15: lc_entry <= 8'h14;
            5'h16: lc_entry <= 8'h60;
            5'h17: lc_entry <= 8'h16;
            5'h18: lc_entry <= 8'hC0;
            5'h19: lc_entry <= 8'h18;
            5'h1A: lc_entry <= 8'h48;
            5'h1B: lc_entry <= 8'h1A;
            5'h1C: lc_entry <= 8'h10;
            5'h1D: lc_entry <= 8'h1C;
            5'h1E: lc_entry <= 8'h20;
            5'h1F: lc_entry <= 8'h1E;
        endcase
    end

    reg [ 7:0] lc;
    reg [23:0] lc_oR;
    reg        lc_reset;
    
    always @( posedge in_hfr_clk ) begin
        if ( lc_reset ) begin
            lc       <= lc_entry;
            lc_oR    <= {reg_2, reg_1, reg_0};
            lc_reset <= 1'b0;
        end
        else begin
            if ( ~lc_halt && lc > 8'd0 ) begin
                lc <= lc - 8'd1;
            end
            if ( lc_oR != {reg_2, reg_1, reg_0} ) begin
                lc_reset <= 1'b1;
            end
        end
    end
    
endmodule
