module pulse (
    input                    in_apu_clk,
    input                    in_qfr_clk,
    input                    in_hfr_clk,
    input             [ 7:0] reg_0,
    input             [ 7:0] reg_1,
    input             [ 7:0] reg_2,
    input             [ 7:0] reg_3,
    output reg signed [15:0] pulse_out,
    output            [ 3:0] test
);

    assign test = swp_prd;

    // reg_0
    wire [1:0] dc_raw;
    wire       lc_halt;
    wire       const_vol;
    wire [3:0] env_prd;
    
    assign dc_raw    = reg_0[7:6];
    assign lc_halt   = reg_0[5];
    assign const_vol = reg_0[4];
    assign env_prd   = reg_0[3:0];
    
    // reg_1
    wire       swp_en;
    wire [2:0] swp_prd;
    wire       swp_neg;
    wire [2:0] swp_shamt;

    assign swp_en    = reg_1[7];
    assign swp_prd   = reg_1[6:4];
    assign swp_neg   = reg_1[3];
    assign swp_shamt = reg_1[2:0];
    
    // reg_2 and reg_3
    wire [10:0] timer;
    wire [ 4:0] length;
    
    assign timer[ 7:0] = reg_2;
    assign timer[10:8] = reg_3[2:0];
    assign length      = reg_3[7:3];
    
    /// Duty cycle logic //////////////////////////////////////////////////////
    reg [7:0] dc;
    
    always @( * )
    begin
        case ( dc_raw )
            2'b00: dc <= 8'b00000010;
            2'b01: dc <= 8'b00000110;
            2'b10: dc <= 8'b00011110;
            2'b11: dc <= 8'b11111001;
        endcase
    end
    
    /// Length counter logic //////////////////////////////////////////////////
    reg [ 7:0] l;
    reg        lc_reset;
    reg [31:0] lc_oR;
    
    always @( * )
    begin
        case ( length )
            5'h00: l <= 8'h0A;
            5'h01: l <= 8'hFE;
            5'h02: l <= 8'h14;
            5'h03: l <= 8'h02;
            5'h04: l <= 8'h28;
            5'h05: l <= 8'h04;
            5'h06: l <= 8'h50;
            5'h07: l <= 8'h06;
            5'h08: l <= 8'hA0;
            5'h09: l <= 8'h08;
            5'h0A: l <= 8'h3C;
            5'h0B: l <= 8'h0A;
            5'h0C: l <= 8'h0E;
            5'h0D: l <= 8'h0C;
            5'h0E: l <= 8'h1A;
            5'h0F: l <= 8'h0E;
            5'h10: l <= 8'h0C;
            5'h11: l <= 8'h10;
            5'h12: l <= 8'h18;
            5'h13: l <= 8'h12;
            5'h14: l <= 8'h30;
            5'h15: l <= 8'h14;
            5'h16: l <= 8'h60;
            5'h17: l <= 8'h16;
            5'h18: l <= 8'hC0;
            5'h19: l <= 8'h18;
            5'h1A: l <= 8'h48;
            5'h1B: l <= 8'h1A;
            5'h1C: l <= 8'h10;
            5'h1D: l <= 8'h1C;
            5'h1E: l <= 8'h20;
            5'h1F: l <= 8'h1E;
        endcase
    end
    
    always @( posedge in_hfr_clk ) begin
        if ( lc_reset ) begin
            len      <= l;
            lc_oR    <= {reg_3, reg_2, reg_1, reg_0};
            lc_reset <= 1'b0;
        end
        else begin
            if ( ~lc_halt && len > 8'd0 ) begin
                len <= len - 8'd1;
            end
            if ( lc_oR != {reg_3, reg_2, reg_1, reg_0} ) begin
                lc_reset <= 1'b1;
            end
        end
    end

    /// Envelope unit /////////////////////////////////////////////////////////
    reg [ 3:0] env_div;
    reg [ 3:0] env_cnt;
    reg [ 3:0] env_out;
    reg        env_start;
    reg [31:0] env_oR;

    always @( posedge in_qfr_clk ) begin

        if ( env_start ) begin
            env_div   <= env_prd;
            env_cnt   <= 4'hf;
            env_oR    <= {reg_3, reg_2, reg_1, reg_0};
            env_start <= 1'b0;
        end
        else begin
            if ( env_div == 4'h0 ) begin
                env_div <= env_prd;

                if ( env_cnt > 4'h0 ) begin
                    env_cnt <= env_cnt - 4'h1;
                end
                else if ( lc_halt ) begin
                    env_cnt <= 4'hf;
                end
            end
            else begin
                env_div <= env_div - 4'h1;
            end

            if ( const_vol ) begin
                env_out <= env_prd;
            end
            else begin
                env_out <= env_cnt;
            end

            if ( env_oR != {reg_3, reg_2, reg_1, reg_0} ) begin
                env_start <= 1'b1;
            end
        end

    end // end always

    /// Sweep unit ////////////////////////////////////////////////////////////
    reg [ 2:0] swp_div;
    reg [10:0] timer_adj;
    reg [31:0] swp_oR;
    reg        swp_reload;

    always @( posedge in_hfr_clk ) begin
        
        if ( swp_reload ) begin
            swp_div    <= swp_prd;
            timer_adj  <= timer;
            swp_oR     <= {reg_3, reg_2, reg_1, reg_0};
            swp_reload <= 1'b0;

            // Adjust pulse channel period
            // Eventually need to check if target period > 0x7ff
            // and other checks as well
            if ( swp_div == 3'd0 && swp_en ) begin
                // Sweep down to lower frequencies
                if ( ~swp_neg ) begin
                    timer_adj <= timer_adj + (timer >> swp_shamt);
                end
                // Sweep up to higher frequencies
                else begin
                    timer_adj <= timer_adj - (timer >> swp_shamt);
                end
            end
        end
        else begin
            if ( swp_div > 3'd0 ) begin
                swp_div <= swp_div - 3'd1;
            end
            else if ( swp_en ) begin
                swp_div <= swp_prd;

                // Adjust pulse channel period
                // Eventually need to implement other checks as well

                // Sweep down to lower frequencies
                if ( ~swp_neg ) begin
                    timer_adj <= timer_adj + (timer >> swp_shamt);
                end
                // Sweep up to higher frequencies
                else begin
                    timer_adj <= timer_adj - (timer >> swp_shamt);
                end
            end

            if ( swp_oR != {reg_3, reg_2, reg_1, reg_0} ) begin
                swp_reload <= 1'b1;
            end
        end

    end // end always

    /// Timer & sequencer /////////////////////////////////////////////////////
    reg [10:0] t;
    reg [ 7:0] len;
    reg [ 2:0] i;
    reg [31:0] seq_oR;
    reg        seq_reset;

    always @( posedge in_apu_clk ) begin
        
        if ( seq_reset ) begin
            i         <= 3'd0;
            t         <= timer_adj;
            seq_oR    <= {reg_3, reg_2, reg_1, reg_0};
            seq_reset <= 1'b0;
        end
        
        if ( t == 11'd0 && len > 8'd0 ) begin
            
            if ( ( dc >> i ) & 8'h1 ) begin
                pulse_out <=  (env_out << 11);
            end
            else begin
                pulse_out <= -(env_out << 11);
            end
            
            if ( i == 0 ) begin
                i <= 3'd7;
            end
            else begin
                i <= i - 3'b001;
            end
            
            t <= timer_adj;
        end
        
        else if ( len > 8'd0 ) begin
            t <= t - 11'd1;
        end
        
        else begin
            if ( seq_oR != {reg_3, reg_2, reg_1, reg_0} ) begin
                seq_reset <= 1'b1;
            end
        end

    end // end always block   
    
endmodule
