module triangle (
    input                    in_nes_clk,
    input                    in_qfr_clk,
    input                    in_hfr_clk,
    input             [ 7:0] reg_0,
    input             [ 7:0] reg_1,
    input             [ 7:0] reg_2,
    output reg signed [15:0] triangle_out,
    input                    reset   
);

    // reg_0
    wire       linc_control;
    wire       lc_halt;
    wire [6:0] linc_reload_val;
    
    assign lc_halt         = reg_0[7];
    assign linc_reload_val = reg_0[6:0];
    
    // reg_1 and reg_2
    wire [10:0] timer;
    wire [ 4:0] lc_load;
    
    assign timer[ 7:0] = reg_1;
    assign timer[10:8] = reg_2[2:0];
    assign lc_load     = reg_2[7:3];

    /// Sequencer logic //////////////////////////////////////////////////
    reg [3:0] seq;

    always @( * )
    begin
        case ( i )
            5'h00: seq <= 4'hf;
            5'h01: seq <= 4'he;
            5'h02: seq <= 4'hd;
            5'h03: seq <= 4'hc;
            5'h04: seq <= 4'hb;
            5'h05: seq <= 4'ha;
            5'h06: seq <= 4'h9;
            5'h07: seq <= 4'h8;
            5'h08: seq <= 4'h7;
            5'h09: seq <= 4'h6;
            5'h0A: seq <= 4'h5;
            5'h0B: seq <= 4'h4;
            5'h0C: seq <= 4'h3;
            5'h0D: seq <= 4'h2;
            5'h0E: seq <= 4'h1;
            5'h0F: seq <= 4'h0;
            5'h10: seq <= 4'h0;
            5'h11: seq <= 4'h1;
            5'h12: seq <= 4'h2;
            5'h13: seq <= 4'h3;
            5'h14: seq <= 4'h4;
            5'h15: seq <= 4'h5;
            5'h16: seq <= 4'h6;
            5'h17: seq <= 4'h7;
            5'h18: seq <= 4'h8;
            5'h19: seq <= 4'h9;
            5'h1A: seq <= 4'ha;
            5'h1B: seq <= 4'hb;
            5'h1C: seq <= 4'hc;
            5'h1D: seq <= 4'hd;
            5'h1E: seq <= 4'he;
            5'h1F: seq <= 4'hf;
        endcase
    end

    /// Linear counter ////////////////////////////////////////////////////////
    reg [23:0] linc_oR;
    reg        linc_reload;
	reg [6:0]  linc;

    always @( posedge in_qfr_clk )
    begin
        if ( linc_reload ) begin
            linc <= linc_reload_val;
            linc_oR <= {reg_2, reg_1, reg_0};
        end
        else if ( linc > 7'd0 ) begin
            linc <= linc - 7'd1;
        end
		
		if ( linc_oR != {reg_2, reg_1, reg_0} ) begin
            linc_reload <= 1'b1;
        end
        else if ( ~lc_halt ) begin
            linc_reload <= 1'b0;
        end
        
    end    
    
    /// Timer & Sequencer /////////////////////////////////////////////////////
    reg [10:0] t;
    reg [ 4:0] i;
    reg [23:0] seq_oR;
    reg        seq_reset;

    always @( posedge in_nes_clk )
    begin
        
        // Reset
        if ( seq_reset ) begin
            seq_reset <= 1'b0;
            t         <= timer;
            seq_oR    <= {reg_2, reg_1, reg_0};
        end

        else begin
        
            // Silence channel if above audible range
				if ( timer < 11'd90 ) begin
					
					triangle_out <= ( ( seq - 4'd8 )  << 12 );
					
				end
		  
            // Sequence through waveform; reload timer
            else if ( t == 11'd0 && lc > 8'd0 && linc > 7'd0 ) begin
                        
                triangle_out <= ( ( seq - 4'd8 )  << 12 );
                
                if ( i == 5'h1F ) begin
                    i <= 0;
                end
                else begin
                    i <= i + 5'd1;
                end
                
                t  <= timer;
            end

            // Decrement counter
            else if ( lc > 8'd0 && linc > 7'd0) begin
                t <= t - 11'd1;
            end

            // Watch for reset
            if ( seq_oR != {reg_2, reg_1, reg_0} ) begin
                seq_reset <= 1'b1;
            end

        end // end else

    end // end always

    /// Length counter logic //////////////////////////////////////////////////
    reg [ 7:0] lc_entry;
    
    always @( * )
    begin
        case ( lc_load )
            5'h00: lc_entry <= 8'h0A;
            5'h01: lc_entry <= 8'hFE;
            5'h02: lc_entry <= 8'h14;
            5'h03: lc_entry <= 8'h02;
            5'h04: lc_entry <= 8'h28;
            5'h05: lc_entry <= 8'h04;
            5'h06: lc_entry <= 8'h50;
            5'h07: lc_entry <= 8'h06;
            5'h08: lc_entry <= 8'hA0;
            5'h09: lc_entry <= 8'h08;
            5'h0A: lc_entry <= 8'h3C;
            5'h0B: lc_entry <= 8'h0A;
            5'h0C: lc_entry <= 8'h0E;
            5'h0D: lc_entry <= 8'h0C;
            5'h0E: lc_entry <= 8'h1A;
            5'h0F: lc_entry <= 8'h0E;
            5'h10: lc_entry <= 8'h0C;
            5'h11: lc_entry <= 8'h10;
            5'h12: lc_entry <= 8'h18;
            5'h13: lc_entry <= 8'h12;
            5'h14: lc_entry <= 8'h30;
            5'h15: lc_entry <= 8'h14;
            5'h16: lc_entry <= 8'h60;
            5'h17: lc_entry <= 8'h16;
            5'h18: lc_entry <= 8'hC0;
            5'h19: lc_entry <= 8'h18;
            5'h1A: lc_entry <= 8'h48;
            5'h1B: lc_entry <= 8'h1A;
            5'h1C: lc_entry <= 8'h10;
            5'h1D: lc_entry <= 8'h1C;
            5'h1E: lc_entry <= 8'h20;
            5'h1F: lc_entry <= 8'h1E;
        endcase
    end

    reg [ 7:0] lc;
    reg [23:0] lc_oR;
    reg        lc_reset;
    
    always @( posedge in_hfr_clk ) begin
        if ( lc_reset ) begin
            lc       <= lc_entry;
            lc_oR    <= {reg_2, reg_1, reg_0};
            lc_reset <= 1'b0;
        end
        else begin
            if ( ~lc_halt && lc > 8'd0 ) begin
                lc <= lc - 8'd1;
            end
            if ( lc_oR != { reg_2, reg_1, reg_0} ) begin
                lc_reset <= 1'b1;
            end
        end
    end
    
endmodule
