module i2c_av_config (
    // Host side
    input  iCLK,
    input  iRST_N,
    
    // I2C side
    inout  I2C_SDAT,
    output I2C_SCLK
);


// Internal Registers/Wires
reg [15:0] mI2C_CLK_DIV;
reg [23:0] mI2C_DATA;
reg        mI2C_CTRL_CLK;
reg        mI2C_GO;
wire       mI2C_END;
wire       mI2C_ACK;

reg [15:0] LUT_DATA;
reg [5:0]  LUT_INDEX;
reg [3:0]  mSetup_ST;

// Clock Setting
parameter CLK_Freq = 50000000; // 50 MHz
parameter I2C_Freq = 20000;  // 20 KHz

// LUT Data Number
parameter LUT_SIZE = 13;

// Audio Data Index
parameter Dummy_DATA  = 0;
parameter SET_LIN_L   = 1;
parameter SET_LIN_R   = 2;
parameter SET_HEAD_L  = 3;
parameter SET_HEAD_R  = 4;
parameter A_PATH_CTRL = 5;
parameter D_PATH_CTRL = 6;
parameter POWER_ON    = 7;
parameter SET_FORMAT  = 8;
parameter SAMPLE_CTRL = 9;
parameter SET_ACTIVE  = 10;
parameter SET_VIDEO   = 11;

// I2C Control Clock //////////////////////////////////////////////////////////
always@(posedge iCLK or negedge iRST_N)
begin
    if(!iRST_N)
    begin
        mI2C_CTRL_CLK <=  1'b0;
        mI2C_CLK_DIV  <= 16'b0;
    end
    else if ( mI2C_CLK_DIV < ( CLK_Freq/I2C_Freq ) )
        mI2C_CLK_DIV  <= mI2C_CLK_DIV + 16'd1;
    else
    begin
        mI2C_CTRL_CLK <= ~mI2C_CTRL_CLK;
        mI2C_CLK_DIV  <= 0;
    end
end

i2c_controller u0 (
    .CLOCK    ( mI2C_CTRL_CLK ),
    .RESET    ( iRST_N        ),
    .I2C_SCLK ( I2C_SCLK      ),
    .I2C_SDAT ( I2C_SDAT      ),
    .I2C_DATA ( mI2C_DATA     ),
    .GO       ( mI2C_GO       ),
    .END      ( mI2C_END      ),
    .ACK      ( mI2C_ACK      )
);
      
// Config Control /////////////////////////////////////////////////////////////
always@( posedge mI2C_CTRL_CLK or negedge iRST_N )
begin
    if ( !iRST_N )
    begin
        LUT_INDEX <= 0;
        mSetup_ST <= 0;
        mI2C_GO  <= 0;
    end
    else if ( LUT_INDEX < LUT_SIZE )
    begin
        case ( mSetup_ST )
            0:
            begin
                if ( LUT_INDEX < SET_VIDEO )
                    mI2C_DATA <= { 8'h34, LUT_DATA };
                else
                    mI2C_DATA <= { 8'h40, LUT_DATA };
     
                mI2C_GO   <= 1;
                mSetup_ST <= 1;
            end
   
            1: 
            begin
                if ( mI2C_END )
                begin
                    if ( !mI2C_ACK )
                        mSetup_ST <= 2;
                    else
                        mSetup_ST <= 0;       
                    
                    mI2C_GO  <= 0;
                end
            end
            
            2:
            begin
                LUT_INDEX <= LUT_INDEX + 6'd1;
                mSetup_ST <= 0;
            end
        endcase
    end
end

// Config data LUT //////////////////////////////////////////////////////////// 
always
begin
    case( LUT_INDEX )
        // Audio Config Data
        Dummy_DATA : LUT_DATA <= 16'h0000;
        SET_LIN_L  : LUT_DATA <= 16'h001f; // Lline in high gain
        SET_LIN_R  : LUT_DATA <= 16'h021f; // Rline in high gain
        SET_HEAD_L : LUT_DATA <= 16'h0479; // Lphone out high vol 7b
        SET_HEAD_R : LUT_DATA <= 16'h0679; // Rphone out high vol
        A_PATH_CTRL: LUT_DATA <= 16'h0810; // Line->ADC, DAC on, no bypass or sidetone 
        D_PATH_CTRL: LUT_DATA <= 16'h0A06; // Deemph to 48kHz
        POWER_ON   : LUT_DATA <= 16'h0C00; // All on
        SET_FORMAT : LUT_DATA <= 16'h0E01; // MSB first left-justified, slave mode
        SAMPLE_CTRL: LUT_DATA <= 16'h1002; // Normal-mode, 384fs oversamp 16'h1002
        SET_ACTIVE : LUT_DATA <= 16'h1201; // Activate
        default    : LUT_DATA <= 16'hxxxx;
    endcase
end

endmodule
