module nes_apu (
    // Clock Input
    input         TD_CLK27,    // 27 MHz
    input         CLOCK_50,    // 50 MHz
    
    // Push Button
    input  [3:0]  KEY,         // Pushbutton[3:0]
	 input  [3:0]  SW,
        
    // SDRAM Interface
    inout  [31:0] DRAM_DQ,     // SDRAM Data bus 32 Bits
    output [12:0] DRAM_ADDR,   // SDRAM Address bus 13 Bits
    output [1:0]  DRAM_BA,     // SDRAM Bank Address
    output [3:0]  DRAM_DQM,    // SDRAM Byte Data Mask 
    output        DRAM_RAS_N,  // SDRAM Row Address Strobe
    output        DRAM_CAS_N,  // SDRAM Column Address Strobe
    output        DRAM_CKE,    // SDRAM Clock Enable
    output        DRAM_CLK,    // SDRAM Clock
    output        DRAM_WE_N,   // SDRAM Write Enable
    output        DRAM_CS_N,   // SDRAM Chip Select
    
    // SD Card
    output        SD_CLK,      // SD Clock
    inout         SD_CMD,      // SD Command
    inout  [3:0]  SD_DAT,      // SD Data
    input         SD_WP_N,     // SD WP
    
    // I2C
    inout         I2C_SDAT,    // I2C Data
    output        I2C_SCLK,    // I2C Clock
    
    // Audio CODEC
    inout         AUD_ADCLRCK, // Audio CODEC ADC LR Clock
    input         AUD_ADCDAT,  // Audio CODEC ADC Data
    inout         AUD_DACLRCK, // Audio CODEC DAC LR Clock
    output        AUD_DACDAT,  // Audio CODEC DAC Data
    inout         AUD_BCLK,    // Audio CODEC Bit-Stream Clock
    output        AUD_XCK,     // Audio CODEC Chip Clock
    
    // TV Decoder
    output        TD_RESET_N,  // TV Decoder Reset
  
    // 7-SEG Display
    output [6:0]  HEX0,        // Seven Segment Digit 0
    output [6:0]  HEX1,        // Seven Segment Digit 1
    output [6:0]  HEX2,        // Seven Segment Digit 2
    output [6:0]  HEX3,        // Seven Segment Digit 3
    output [6:0]  HEX4,        // Seven Segment Digit 4
    output [6:0]  HEX5,        // Seven Segment Digit 5
    output [6:0]  HEX6,        // Seven Segment Digit 6
    output [6:0]  HEX7         // Seven Segment Digit 7
);
    
    /// Nios System ///////////////////////////////////////////////////////////
    wire [7:0] apu_addr;
    wire [7:0] apu_data;
    wire [7:0] apu_wren;
    wire       frame_intr;
    
    reg  [7:0] apu_reg [32];
    
    wire nios_clk;
    
    nios cpu(
        .clk_clk          ( nios_clk   ),
        .sdram_addr       ( DRAM_ADDR  ),
        .sdram_ba         ( DRAM_BA    ),
        .sdram_cas_n      ( DRAM_CAS_N ),
        .sdram_cke        ( DRAM_CKE   ),
        .sdram_cs_n       ( DRAM_CS_N  ),
        .sdram_dq         ( DRAM_DQ    ),
        .sdram_dqm        ( DRAM_DQM   ),
        .sdram_ras_n      ( DRAM_RAS_N ),
        .sdram_we_n       ( DRAM_WE_N  ),
        .sd_clk_export    ( SD_CLK     ),
        .sd_dat_export    ( SD_DAT     ),
        .sd_cmd_export    ( SD_CMD     ),
        .sd_wp_n_export   ( SD_WP_N    ),
        .apu_addr_export  ( apu_addr   ),
        .apu_data_export  ( apu_data   ),
        .apu_wren_export  ( apu_wren   ),
        .key_0_export     ( KEY[0]     ),
        .key_1_export     ( KEY[1]     ),
        .key_2_export     ( KEY[2]     ),
        .key_3_export     ( KEY[3]     )
    );
    
    sdram_pll neg_3ns (
        .inclk0( CLOCK_50 ), 
        .c0    ( DRAM_CLK ), 
        .c1    ( nios_clk )
    );
    
    //nios_pll neg_3ns (
    //    .inclk0( CLOCK_50 ), 
    //    .c0    ( DRAM_CLK ), 
    //    .c1    ( nios_clk )
    //);
    
    /// APU Reg Logic /////////////////////////////////////////////////////////
    always @( * )
    begin
        if ( apu_wren )
            apu_reg[apu_addr] <= apu_data;
    end
    
    /// APU ///////////////////////////////////////////////////////////////////
    wire nes_clk;
    wire apu_clk;
    wire qfr_clk;
    wire hfr_clk;

    wire [13:0] fc_test;
    wire [ 3:0] p_test;
    
    wire signed [15:0] p1_out;
    wire signed [15:0] p2_out;
    wire signed [15:0] t_out;
    wire signed [15:0] n_out;
    
    nes_pll pll_1 (
        .inclk0 ( TD_CLK27 ),
        .c0     ( nes_clk  ),
        .c1     ( apu_clk  )
    );
    
    frame_counter fc (
        .clk           ( apu_clk    ),
        .fc_qfr_clk    ( qfr_clk    ),
        .fc_hfr_clk    ( hfr_clk    ),
        .fc_frame_intr ( frame_intr ),
        .test          ( fc_test    )
    );
    
    pulse p1 (
        .in_apu_clk ( apu_clk      ),
        .in_qfr_clk ( qfr_clk      ),
        .in_hfr_clk ( hfr_clk      ),
        .reg_0      ( apu_reg[0]   ),
        .reg_1      ( apu_reg[1]   ),
        .reg_2      ( apu_reg[2]   ),
        .reg_3      ( apu_reg[3]   ),
        .pulse_out  ( p1_out       ),
        .test       ( p_test       )
    );
    
    pulse p2 (
        .in_apu_clk ( apu_clk    ),
        .in_qfr_clk ( qfr_clk    ),
        .in_hfr_clk ( hfr_clk    ),
        .reg_0      ( apu_reg[4] ),
        .reg_1      ( apu_reg[5] ),
        .reg_2      ( apu_reg[6] ),
        .reg_3      ( apu_reg[7] ),
        .pulse_out  ( p2_out     )
    );
    
    triangle t (
        .in_nes_clk   ( nes_clk     ),
        .in_qfr_clk   ( qfr_clk     ),
        .in_hfr_clk   ( hfr_clk     ),
        .reg_0        ( apu_reg[8]  ),
        .reg_1        ( apu_reg[10] ),
        .reg_2        ( apu_reg[11] ),
        .triangle_out ( t_out       )
    );
    
    noise n (
        .in_apu_clk ( apu_clk     ),
        .in_qfr_clk ( qfr_clk     ),
        .in_hfr_clk ( hfr_clk     ),
        .reg_0      ( apu_reg[12] ),
        .reg_1      ( apu_reg[14] ),
        .reg_2      ( apu_reg[15] ),
        .noise_out  ( n_out       )
    );
    
    /// Audio Driver //////////////////////////////////////////////////////////
    wire   reset;
    assign reset = ~KEY[0];
    
    wire audio_clk;
    wire dly_rst;
    
    wire signed [15:0] audio_outL;
    wire signed [15:0] audio_outR;

    assign TD_RESET_N  = 1'b1;
    assign AUD_ADCLRCK = AUD_DACLRCK;
    assign AUD_XCK     = audio_clk;
    
	 wire signed [15:0] a1;
	 wire signed [15:0] a2;
	 wire signed [15:0] a3;
	 wire signed [15:0] a4;
	 
	 assign a1 = SW[0] ? (p1_out >>> 3) : 16'b0;
	 assign a2 = SW[1] ? (p2_out >>> 3) : 16'b0;
	 assign a3 = SW[2] ? ( t_out >>> 3) : 16'b0;
	 assign a4 = SW[3] ? ( n_out >>> 3) : 16'b0;
	 
    //assign audio_outL = a1 + a2 + a3 + a4 ;
    assign audio_outL = audio_outR;
    assign audio_outR = (p1_out >>> 3) + (p2_out >>> 3) + (n_out >>> 3) + (t_out >>> 3);

    reset_delay r0 (
        .iCLK   ( CLOCK_50 ),
        .oRESET ( dly_rst  )
    );

    audio_pll pll (
        .areset ( ~dly_rst  ),
        .inclk0 ( TD_CLK27  ),
        .c0     ( audio_clk )
    );

    i2c_av_config u3 (
        .iCLK     ( CLOCK_50 ),
        .iRST_N   ( ~reset   ),
        .I2C_SCLK ( I2C_SCLK ),
        .I2C_SDAT ( I2C_SDAT )
    );

    audio_dac_adc u4 (
        .oAUD_BCK    ( AUD_BCLK    ),
        .oAUD_DATA   ( AUD_DACDAT  ),
        .oAUD_LRCK   ( AUD_DACLRCK ),
        .iAUD_ADCDAT ( AUD_ADCDAT  ),
        .iAUD_extL   ( audio_outL  ),
        .iAUD_extR   ( audio_outR  ),
        .iCLK_18_4   ( audio_clk   ),
        .iRST_N      ( dly_rst     )
    );
    
    /// Hex Displays //////////////////////////////////////////////////////////
   
    hex hd7 (
       .value ( apu_reg[2][7:4] ),
       .disp  ( HEX7 )
    );
    
    hex hd6 (
       .value ( apu_reg[2][3:0] ),
       .disp  ( HEX6 )
    );
 
    
    hex hd5 (
       .value ( apu_reg[6][7:4] ),
       .disp  ( HEX5 )
    );
    
    hex hd4 (
       .value ( apu_reg[6][3:0] ),
       .disp  ( HEX4 )
    );
    
    hex hd3 (
       .value ( apu_reg[10][7:4]  ),
       .disp  ( HEX3 )
    );
    
    hex hd2 (
       .value ( apu_reg[10][3:0] ),
       .disp  ( HEX2 )
    );
    
    hex hd1 (
       .value ( apu_reg[14][7:4] ),
       .disp  ( HEX1 )
    );
    
    hex hd0 (
       .value ( apu_reg[14][3:0] ),
       .disp  ( HEX0 )
    );

    /*
    hex hd7 (
       .value ( 4'hf ),
       .disp  ( HEX7 )
    );
    
    hex hd6 (
       .value ( {3'b0, frame_intr} ),
       .disp  ( HEX6 )
    );
 
    hex hd5 (
       .value ( 4'h0 ),
       .disp  ( HEX5 )
    );
    
    hex hd4 (
       .value ( {3'b0, qfr_clk} ),
       .disp  ( HEX4 )
    );
 
    hex hd3 (
       .value ( fc_test[13:12] ),
       .disp  ( HEX3 )
    );
    
    hex hd2 (
       .value ( fc_test[11: 8] ),
       .disp  ( HEX2 )
    );
    
    hex hd1 (
       .value ( fc_test[ 7: 4] ),
       .disp  ( HEX1 )
    );
    
    hex hd0 (
       .value ( fc_test[ 3: 0] ),
       .disp  ( HEX0 )
    );
    */
    
endmodule
